library IEEE;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.NUMERIC_STD.ALL;

entity SineGen is
    port(CLOCK_50 :   in std_logic;
         VGAout:      out unsigned(10 downto 0); -- r2,r1,r0, g2,g1,g0, b2,b1,b0, hsync, vsync
         PS2_Clock:   in std_logic; 
         PS2_Data:    in std_logic
         );
end SineGen;

architecture Behavioral of SineGen is

type charmaparray               is array(15 downto 0) of unsigned(63 downto 0);

constant Red                                   : unsigned(8 downto 0):="111000000";
constant Green                                 : unsigned(8 downto 0):="000111000";
constant Blue                                  : unsigned(8 downto 0):="000000111";
constant Yellow                                : unsigned(8 downto 0):="111111000";
constant Cyan                                  : unsigned(8 downto 0):="000111111";
constant White                                 : unsigned(8 downto 0):="111111111";
constant Black                                 : unsigned(8 downto 0):="000000000";
signal hcount                                  : unsigned(10 downto 0):="00000000000";
signal vcount                                  : unsigned(9 downto 0):="0000000000";
signal scancount                               : unsigned(10 downto 0):="00000000000";
signal Yval                                    : unsigned(9 downto 0):="0000000000";
signal VGA, Out_Colour                         : unsigned(8 downto 0):="000000000";
signal videoon, videov, videoh, hsync, vsync   : std_ulogic:='0';
signal scancountdir                            : std_ulogic:='0';
signal reset                                   : std_ulogic:='0';
signal fraccountdir                            : std_ulogic:='1';
signal Red_Pixel, Green_Pixel, Blue_Pixel      : std_ulogic:='0';
signal Yellow_Pixel, White_Pixel, Black_Pixel  : std_ulogic:='0';
signal Cyan_Pixel                              : std_ulogic:='0';
signal aa, bb                                  : unsigned(7 downto 0);
signal aIN, bIN                                : unsigned(3 downto 0):="0000";
signal pOUT                                    : unsigned(7 downto 0):="00000000";

constant char_0                                : unsigned(63 downto 0):="00000000"
                                                                       &"01111100"
                                                                       &"10000110"
                                                                       &"10001010"
                                                                       &"10010010"
                                                                       &"10100010"
                                                                       &"11000010"
                                                                       &"01111100";

constant char_1                                : unsigned(63 downto 0):="00000000"
                                                                       &"00010000"
                                                                       &"00110000"
                                                                       &"01010000"
                                                                       &"00010000"
                                                                       &"00010000"
                                                                       &"00010000"
                                                                       &"11111110";

constant char_2                                : unsigned(63 downto 0):="00000000"
                                                                       &"01111100"
                                                                       &"10000010"
                                                                       &"00000010"
                                                                       &"01111100"
                                                                       &"10000000"
                                                                       &"10000010"
                                                                       &"11111110";

constant char_3                                : unsigned(63 downto 0):="00000000"
                                                                       &"01111100"
                                                                       &"10000010"
                                                                       &"00000010"
                                                                       &"00011100"
                                                                       &"00000010"
                                                                       &"10000010"
                                                                       &"01111100";

constant char_4                                : unsigned(63 downto 0):="00000000"
                                                                       &"00001000"
                                                                       &"00011000"
                                                                       &"00101000"
                                                                       &"01001000"
                                                                       &"11111110"
                                                                       &"00001000"
                                                                       &"00011100";

constant char_5                                : unsigned(63 downto 0):="00000000"
                                                                       &"11111110"
                                                                       &"10000010"
                                                                       &"10000000"
                                                                       &"01111100"
                                                                       &"00000010"
                                                                       &"10000010"
                                                                       &"01111100";

constant char_6                                : unsigned(63 downto 0):="00000000"
                                                                       &"01111100"
                                                                       &"10000010"
                                                                       &"10000000"
                                                                       &"11111100"
                                                                       &"10000010"
                                                                       &"10000010"
                                                                       &"01111100";

constant char_7                                : unsigned(63 downto 0):="00000000"
                                                                       &"11111110"
                                                                       &"00000100"
                                                                       &"00000100"
                                                                       &"00111000"
                                                                       &"00001000"
                                                                       &"00010000"
                                                                       &"00010000";

constant char_8                                : unsigned(63 downto 0):="00000000"
                                                                       &"01111100"
                                                                       &"10000010"
                                                                       &"10000010"
                                                                       &"01111100"
                                                                       &"10000010"
                                                                       &"10000010"
                                                                       &"01111100";

constant char_9                                : unsigned(63 downto 0):="00000000"
                                                                       &"01111100"
                                                                       &"10000010"
                                                                       &"10000010"
                                                                       &"01111110"
                                                                       &"00000010"
                                                                       &"10000010"
                                                                       &"01111100";

constant char_A                                : unsigned(63 downto 0):="00000000"
                                                                       &"00010000"
                                                                       &"00101000"
                                                                       &"00101000"
                                                                       &"00111000"
                                                                       &"01000100"
                                                                       &"01000100"
                                                                       &"11101110";

constant char_B                                : unsigned(63 downto 0):="00000000"
                                                                       &"11111000"
                                                                       &"01000100"
                                                                       &"01111100"
                                                                       &"01000010"
                                                                       &"01000010"
                                                                       &"01000010"
                                                                       &"11111100";

constant char_C                                : unsigned(63 downto 0):="00000000"
                                                                       &"00111100"
                                                                       &"01000010"
                                                                       &"10000000"
                                                                       &"10000000"
                                                                       &"10000000"
                                                                       &"01000010"
                                                                       &"00111100";

constant char_D                                : unsigned(63 downto 0):="00000000"
                                                                       &"11111000"
                                                                       &"01000100"
                                                                       &"01000010"
                                                                       &"01000010"
                                                                       &"01000010"
                                                                       &"01000100"
                                                                       &"11111000";

constant char_E                                : unsigned(63 downto 0):="00000000"
                                                                       &"11111110"
                                                                       &"01000010"
                                                                       &"01000000"
                                                                       &"01111000"
                                                                       &"01000000"
                                                                       &"01000010"
                                                                       &"11111110";

constant char_F                                : unsigned(63 downto 0):="00000000"
                                                                       &"11111110"
                                                                       &"01000010"
                                                                       &"01000000"
                                                                       &"01111000"
                                                                       &"01000000"
                                                                       &"01000000"
                                                                       &"11100000";

constant char_ANDt                             : unsigned(63 downto 0):="01111111"
                                                                       &"01000000"
                                                                       &"01000000"
                                                                       &"01000000"
                                                                       &"01000000"
                                                                       &"01000000"
                                                                       &"01000000"
                                                                       &"01000000";

constant char_ANDb                             : unsigned(63 downto 0):="01000000"
                                                                       &"01000000"
                                                                       &"01000000"
                                                                       &"01000000"
                                                                       &"00100000"
                                                                       &"00010000"
                                                                       &"00001111"
                                                                       &"00000001";

constant char_plus                             : unsigned(63 downto 0):="00011000"
                                                                       &"00011000"
                                                                       &"00011000"
                                                                       &"11111111"
                                                                       &"11111111"
                                                                       &"00011000"
                                                                       &"00011000"
                                                                       &"00011000";

constant char_times                            : unsigned(63 downto 0):="00000000"
                                                                       &"00000000"
                                                                       &"01000010"
                                                                       &"00100100"
                                                                       &"00011000"
                                                                       &"00011000"
                                                                       &"00100100"
                                                                       &"01000010";

constant char_equals                           : unsigned(63 downto 0):="00000000"
                                                                       &"00000000"
                                                                       &"00000000"
                                                                       &"11111111"
                                                                       &"00000000"
                                                                       &"11111111"
                                                                       &"00000000"
                                                                       &"00000000";

constant char_H                                : unsigned(63 downto 0):="11000011"
                                                                       &"11000011"
                                                                       &"11000011"
                                                                       &"11111111"
                                                                       &"11111111"
                                                                       &"11000011"
                                                                       &"11000011"
                                                                       &"11000011";

                                                                                                        
constant CharMap: charmaparray:=(char_F,char_E,char_D,char_C,char_B,char_A,char_9,char_8,char_7,char_6,char_5,char_4,char_3,char_2,char_1,char_0);

function BinToBCD(r:unsigned) return unsigned;
function v4(a:unsigned;b:unsigned) return unsigned;
function v8(aa:unsigned;bb:unsigned) return unsigned;

function BinToBCD(r:unsigned) return unsigned is
--S variables are intermediate stages in Bin2BCD logic
variable S1,S2,S3,S4,S5,S6,S7   : unsigned(3 downto 0):="0000";

begin
-- 8 bit binary to 12 bit BCD encoder
        if '0'&r(7 downto 5) > "0100" then
            S1 := unsigned('0'&r(7 downto 5)) + 3;
        else
            S1 := unsigned('0'&r(7 downto 5));
        end if;
        if S1(2 downto 0)&r(4) > "0100" then
            S2 := S1(2 downto 0)&r(4) +3;
        else
            S2 := S1(2 downto 0)&r(4);
        end if;
        if S2(2 downto 0)&r(3) > "0100" then
            S3 := S2(2 downto 0)&r(3) +3;
        else
            S3 := S2(2 downto 0)&r(3);
        end if;
        if S3(2 downto 0)&r(2) > "0100" then
            S4 := S3(2 downto 0)&r(2) +3;
        else
            S4 := S3(2 downto 0)&r(2);
        end if;
        if S4(2 downto 0)&r(1) > "0100" then
            S5 := S4(2 downto 0)&r(1) +3;
        else
            S5 := S4(2 downto 0)&r(1);
        end if;
        if S1(3)&S2(3)&S3(3) > 4 then
            S6 := '0'&S1(3)&S2(3)&S3(3) +3;
        else
            S6 := '0'&S1(3)&S2(3)&S3(3);
        end if;
        if S6(2 downto 0)&S4(3) > "0100" then
            S7 := S6(2 downto 0)&S4(3) +3;
        else
            S7 := S6(2 downto 0)&S4(3);
        end if;
    return "00"&S6(3)&S7(3 downto 0)&S5(3 downto 0)&r(0);
end BinToBCD;

function v4(a:unsigned;b:unsigned) return unsigned is
variable r0,r7                  : std_ulogic:='0';
variable r2,r3,r4,r5,r6,t2      : unsigned(2 downto 0):="000";
variable r1,t3,t4,t5            : unsigned(1 downto 0):="00";
variable mid                    : unsigned(15 downto 0):="0000000000000000";

begin
        mid(0)  := a(0) and b(0);
        mid(1)  := a(1) and b(0);
        mid(2)  := a(0) and b(1);
        mid(3)  := a(2) and b(0);
        mid(4)  := a(1) and b(1);
        mid(5)  := a(0) and b(2);
        mid(6)  := a(3) and b(0);
        mid(7)  := a(2) and b(1);
        mid(8)  := a(1) and b(2);
        mid(9)  := a(0) and b(3);
        mid(10) := a(3) and b(1);
        mid(11) := a(2) and b(2);
        mid(12) := a(1) and b(3);
        mid(13) := a(3) and b(2);
        mid(14) := a(2) and b(3);
        mid(15) := a(3) and b(3);
        r0:=mid(0);
        r1(0):=mid(1) xor mid(2);      
        r1(1):=mid(1) and mid(2);
        r2(0):=mid(3) xor mid(4) xor mid(5) xor r1(1);
        r2(1):=(mid(3) and mid(4)) xor (mid(5) and r1(1)) xor (mid(3) and mid(5)) xor (mid(4) and r1(1)) xor (mid(3) and r1(1)) xor (mid(4) and mid(5));
        r2(2):=mid(3) and mid(4) and mid(5) and r1(1);
        t2(0):=mid(6) xor mid(7) xor mid(8) xor mid(9);
        t2(1):=(mid(6) and mid(7)) xor (mid(8) and mid(9)) xor (mid(6) and mid(8)) xor (mid(7) and mid(9)) xor (mid(6) and mid(9)) xor (mid(7) and mid(8));
        t2(2):=mid(6) and mid(7) and mid(8) and mid(9);
        r3 := t2 + to_integer(r2(2 downto 1));
        t3(0):=mid(10) xor mid(11) xor mid(12);
        t3(1):=(mid(10) and mid(11)) or (mid(11) and mid(12)) or (mid(12) and mid(10));
        r4 := '0'&t3 + to_integer(r3(2 downto 1));
        t4(0):=mid(13) xor mid(14);      
        t4(1):=mid(13) and mid(14);
        r5 := '0'&t4 + to_integer(r4(2 downto 1));
        t5(0) :=mid(15);
        r6 := '0'&t5 + to_integer(r5(2 downto 1));
        r7 := r6(1);
    return r7&r6(0)&r5(0)&r4(0)&r3(0)&r2(0)&r1(0)&r0;       
end v4;

function v8(aa:unsigned;bb:unsigned) return unsigned is
variable pt1,pt3                : unsigned(7 downto 0);
variable pt2                    : unsigned(8 downto 0);
variable aal,aah,bbl,bbh    : unsigned(3 downto 0);

begin
aal := aa(3 downto 0);
aah := aa(7 downto 4);
bbl := bb(3 downto 0);
bbh := bb(7 downto 4);
pt1 := v4(aal,bbl);
pt2 := '0'&v4(aal,bbh)+v4(aah,bbl)+pt1(7 downto 4);
pt3 := v4(aah,bbh)+ pt2(8 downto 4);
return pt3&pt2(3 downto 0)&pt1(3 downto 0);
end v8;

begin
hcounter: process (CLOCK_50, reset)
begin
   if reset='1' then
      hcount <= "0000000000";
    elsif (rising_edge(CLOCK_50)) then
      hcount <= hcount + 1;
      if hcount=1039 then
         hcount <= "00000000000";
        end if;
    end if;
end process;

vcounter: process (CLOCK_50, reset)
begin
   if reset='1' then 
      vcount <= "0000000001";
    elsif (rising_edge(CLOCK_50)) then
      if hcount = 1039 then
         vcount <= vcount + 1;
         if vcount = 665 then
                vcount <= "0000000000";
                scancount <= scancount +1;
                if scancount = 640 then
                    scancount <= "00010000000";
                end if;
            end if;
      end if;
   end if;
end process;

process (vcount)
begin
   videov <= '1'; 
   if vcount > 599 then
        videov <= '0';
   end if;
end process;

process (hcount)
begin
   videoh <= '1';
   if hcount > 799 then
        videoh <= '0';
   end if;
end process;

sync: process (CLOCK_50, reset)
begin
   if reset='1'  then 
      hsync <= '0';
      vsync <= '0';
    elsif (rising_edge(CLOCK_50)) then
      hsync <= '1';
      if (hcount <= 987 and hcount >= 855) then
         hsync <= '0';
      end if;
      vsync <= '1';
      if (vcount <= 645 and vcount >= 636) then
         vsync <= '0';
      end if;
   end if;
end process;

PixelColourSelector: process (CLOCK_50,reset)
begin
    if reset='1'  then
        Out_Colour <= Black;
    elsif (rising_edge(CLOCK_50)) then
        Out_Colour <= Black;
        if Green_Pixel='1' then
            Out_Colour <= Green;
        end if;
        if Blue_Pixel='1' then
            Out_Colour <= Blue;
        end if;
        if Yellow_Pixel='1' then
            Out_Colour <= Yellow;
        end if;
        if Cyan_Pixel='1' then
            Out_Colour <= Cyan;
        end if;
        if White_Pixel='1' then
            Out_Colour <= White;
        end if;
        if Red_Pixel='1' then
            Out_Colour <= Red;
        end if;
        if Black_Pixel='1' then
            Out_Colour <= Black;
        end if;
    end if;
end process;

DrawApp: process (CLOCK_50, reset)

variable tmp1           : unsigned(15 downto 0):="0000000000000000";
variable tmp2           : unsigned(15 downto 0):="0000000000000000";
variable pixelh         : unsigned(10 downto 0):="00000000000";
variable pixelv         : unsigned( 9 downto 0):="0000000000";
variable xytemp1        : unsigned( 7 downto 0):="00000000";
variable xytemp2        : unsigned( 7 downto 0):="00000000";
variable BCDTempVar     : unsigned(11 downto 0):="000000000000";
variable Ones           : unsigned(3 downto 0):="0000";
variable Tens           : unsigned(3 downto 0):="0000";
variable Hundreds       : unsigned(3 downto 0):="0000";

begin
    if (rising_edge(CLOCK_50)) then 
            Red_Pixel <= '0';
            Green_Pixel <= '0';
            Blue_Pixel <= '0';
            Yellow_Pixel <= '0';
            Cyan_Pixel <= '0';
            White_Pixel <='0';
            Black_Pixel <='0';

            if Yellow_Pixel = '1' and Green_Pixel = '1' then
            Red_Pixel <='1';
            Yval <= vcount;
            end if;

pixelv := vcount - 000;
pixelh := hcount - 000;
    if pixelv >= 0 and pixelv <= 255 and pixelh >= 0 and pixelh <= 255 then
        xytemp1 := not  pixelh(7 downto 0);
        xytemp2 := not  pixelv(7 downto 0);
    end if;
    if pixelv >= 0 and pixelv <= 255 and pixelh >= 256 and pixelh <= 511 then
        xytemp1 :=      pixelh(7 downto 0);
        xytemp2 := not  pixelv(7 downto 0);
    end if;

pixelv := vcount - 000;
pixelh := hcount - 256;
    if pixelv >= 256 and pixelv <= 511 and pixelh >= 0 and pixelh <= 255 then
        xytemp1 := not  pixelh(7 downto 0);
        xytemp2 :=      pixelv(7 downto 0);
    end if;
    if pixelv >= 256 and pixelv <= 511 and pixelh >= 256 and pixelh <= 511 then
        xytemp1 :=      pixelh(7 downto 0);
        xytemp2 :=      pixelv(7 downto 0);
    end if;
        

    tmp1 := v8(xytemp1 , xytemp1) / 256;
    tmp2 := v8(xytemp2 , xytemp2) / 256;
    if tmp1+tmp2 = 63 then
        Yellow_Pixel <='1';
    end if;
    if ((hcount = 127 or hcount = 640) or (vcount = 127 or vcount = 384)) and
        (hcount >= 127 and hcount <= 640) and (vcount >= 127 and vcount <= 384) then
        White_Pixel <='1';
    end if;
    if ((hcount = scancount) and (vcount > 127 and vcount < 384)) then
        Green_Pixel <='1';
    end if;
    if ((hcount > 127 and hcount < 640) and (vcount > 127 and vcount < 384) and vcount = Yval)  then
        Green_Pixel <='1';
    end if;     

-- Display Y position in decimal digits

BCDTempVar  := BinToBCD(Yval-128);
Ones            := BCDTempVar(3 downto 0);
Tens            := BCDTempVar(7 downto 4);
Hundreds        := BCDTempVar(11 downto 8);

        pixelv := vcount - Yval;
        pixelh := hcount - 671;

        if pixelv >= "0000000000" and pixelv < "0000001000" and pixelh >= "0000000000" and pixelh < "0000001000" and Hundreds > "0000" then
            if CharMap(to_integer(Hundreds))(to_integer(not pixelv(2 downto 0)&not pixelh(2 downto 0)))='1' then
                Cyan_Pixel <='1';
            end if;
        end if;
        if pixelv >= "0000000000" and pixelv < "0000001000" and pixelh >= "0000001000" and pixelh < "0000010000" and (Tens > "0000" or (Tens = "0000" and Hundreds >"0000")) then
            if CharMap(to_integer(Tens))(to_integer(not pixelv(2 downto 0)&not pixelh(2 downto 0)))='1' then
                Cyan_Pixel <='1';
            end if;
        end if;
        if pixelv >= "0000000000" and pixelv < "0000001000" and pixelh >= "0000010000" and pixelh < "0000011000" then
            if CharMap(to_integer(Ones))(to_integer(not pixelv(2 downto 0)&not pixelh(2 downto 0)))='1' then
                Cyan_Pixel <='1';
            end if;
        end if;


-- Draw border
--      White_Pixel <='0';
        if vcount <= 8 or vcount >= 592 or
            hcount <= 8 or hcount >= 792 then
            White_Pixel <='1';
        end if;
    end if;
    
end process;

   videoon                  <= videoh and videov;
    VGA                     <= Out_Colour;
   VGAout(10 downto 2)  <= VGA and videoon&videoon&videoon&videoon&videoon&videoon&videoon&videoon&videoon;
    VGAout(1 downto 0)  <= hsync & vsync;

end Behavioral;

